/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model;

import javax.persistence.AccessType;
import javax.persistence.metamodel.Attribute.PersistentAttributeType;


import ws.ladder.model.event.CollectionEvent;
import ws.ladder.util.ObjectCreator;

public interface ExtendedProperty extends Property {
	/**
	 * Get the access type (if any) specified on the property/field
	 * @return
	 */
	public AccessType getAccessType();

	/**
	 * Use reflection to get the value of this property from the dataObject
	 * @param dataObject
	 * @return
	 */
	public Object getValue(Object dataObject);

	/**
	 * Use reflection to set the value of this property on the dataObject
	 * @param dataObject
	 * @param propertyValue
	 */
	public void setValue(Object dataObject, Object propertyValue);

	/**
	 * Make a shallow copy of the value of this property 
	 * @param dataObject
	 * @return
	 * @throws Exception 
	 */
	public Object shallowCopy(Object dataObject, ObjectCreator creator);

	/**
	 * Gets the key type for a map property 
	 * @return
	 */
	public Type getKeyType();

	/**
	 * Gets the element type for a collection property
	 */
	public Type getElementType();	

	/**
	 * Indicates if this property needs to be initialized when the object is created
	 */
	public boolean needsInitialization();

	/**
	 * Allow business logic to be invoked when adding an object to a collection
	 * @param dataObject
	 * @param propertyValue
	 */
	void setCustomAdd(Object dataObject, CollectionEvent event);

	/**
	 * Returns the type of association modelled by this property
	 * @return
	 */
	public PersistentAttributeType getAssociationType();

	/**
	 * For the  derived property it checks if the XmlTransient annotation is present
	 * For a POJO property it checks if the java transient annotation is present
	 * @return
	 */
	public boolean isTransient();

	/**
	 * Does this property refer to a bi-directional relationship
	 */
	public boolean isBiDirectional();
	
	/**
	 * If this property refer to a bi-directional relationship and the types are symmetrical, i.e., the backRef points to the same type and not to a sub-type
	 */
	public boolean isSymmetricalBiDirectionalType();	

	/**
	 * Returns the mapped by Property of a bi-directional relation. The mapped by property is the property that stores the value of the link in the database.
	 * @return 
	 */
	public Property getMappedBy();

	/**
	 * Returns the Property that is the other side of a bi-directional relation.
	 * @return 
	 */
	public Property getMapOf();

	/**
	 * Allows to work with embedded objects
	 * @param mapPath
	 */
	public void setMapPath(String mapPath);

	/**
	 * Returns the path of the mappedBy property
	 * @return
	 */
	public String getMapPath();
	
	/**
	 * Returns true if the collection represents a Map
	 */
	public boolean isMap();
	
	/**
	 * Returns true if the collection represents a List
	 */
	public boolean isList();	
	
	/**
	 * Returns true if the collection represents a Set
	 */
	public boolean isSet();

	/**
	 * Allow business logic to be invoked when removing an object from a collection
	 * @param dataObject
	 * @param event
	 */	
	void setCustomRemove(Object dataObject, CollectionEvent event);

	/**
	 * Convenience method to see if the property refers to a DataObject or a simple object
	 * This method also considers a collection of DataObject as a DataObject whereas the Type.isDataType does 
	 * not make this difference since it does not have access to this information.
	 * @return
	 */
	boolean isDataType();

	/**
	 * Returns the property on the entity responsible for holding the index/key value for this toMany property
	 * @return
	 */
	public Property getPositionProperty();

	/**
	 * @param property refers to the toMany property which has a map key tracked at this property
	 */
	public void setMapKeyOf(ExtendedProperty property);
	
	/**
	 * @param property refers to the toMany property which has an index tracked at this property
	 */	
	public void setIndexOf(ExtendedProperty property);	
}
