/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model.action;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import ws.ladder.model.AggregateAction;
import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.operation.AbstractOperation;


public class MapUpdateAction extends CollectionUpdateAction {

	private PropertyKey key;
	private CallInfo callInfo;
	private Set<Object> emptyKeys = new HashSet<Object>();
	private Map<Object, Position> positions = new HashMap<Object, Position>(); // position by object id

	public MapUpdateAction(PropertyKey key) {
		this.key = key;
	}

	@Override
	public void execute() {
		// Remove all the keys from the target map not present in the Source Map
		Map targetMap = (Map) callInfo.getTargetFromParent();
		Map sourceMap = (Map) callInfo.getSourceFromParent();
		
		if(sourceMap == null)
			sourceMap = new HashMap();
		for(Object key: targetMap.keySet()) 
			if(!sourceMap.containsKey(key))
				targetMap.remove(key);
		
		// TODO: Invoke RemoveElementAction, AddElementAction and RepositionElementAction
		// Any incomplete actions need to be resolved or an error raised
		// An action is incomplete 
		// 1. if the position is null or
		// 2. position is not null (duplicates will not be created) and the collection is a list or a map
		// 3. setModified flag on map data object

		
		// add the emptyKeys
		for(Object key: emptyKeys)
			targetMap.put(key, null);
	}

	@Override
	public void addAction(Executable action) {
		// TODO Auto-generated method stub

	}

	@Override
	public PropertyKey getKey() {
		return key;
	}

	protected void trackPosition(CallInfo callInfo) {
		// NOTE: if key is a composite (embedded type), then we will need to 
		// invoke processValue on the keys to convert it to an object of the REFERENCE type
		// Right now, the processing is based on the assumption the key is of simple type
		// Also, the identifier is considered to be a simple type

		// track new position
		Map sourceMap = (Map) callInfo.getSourceFromParent();
		if(sourceMap != null) {
			for(Object entry: sourceMap.entrySet()) {
				Object key = ((Map.Entry)entry).getKey();
				Object value = ((Map.Entry)entry).getValue();
				if(value == null) { // the key is set to null
					emptyKeys.add(key);
					continue;
				}
					
				ExtendedDataObject valueDO = (ExtendedDataObject) callInfo.getSourceObjectCreator().createTarget(callInfo, value, null);

				Position position = positions.get(valueDO.getIdentifierValue());
				if(position == null) {
					position = new MapPosition(valueDO);
					positions.put(valueDO.getIdentifierValue(), position);
				}
				position.addToNew(position);
			}
		}

		// track old position
		Map targetMap = (Map) callInfo.getTargetFromParent();
		for(Object entry: targetMap.entrySet()) {
			Object key = ((Map.Entry)entry).getKey();
			ExtendedDataObject value = (ExtendedDataObject) callInfo.getTargetObjectCreator().createTarget(callInfo, ((Map.Entry)entry).getValue(), null);

			Position position = positions.get(value.getIdentifierValue());
			if(position == null) {
				position = new MapPosition(value);
				positions.put(value.getIdentifierValue(), position);
			}
			position.addToOld(position);
		}		
	}

	@Override
	protected void removeCollectionElement(CallInfo callInfo, ExtendedDataObject collectionElement) {
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getTarget();
		
		Position pos = positions.get(collectionElement.getIdentifierValue());
		RemoveElementAction originalAction = null;
		for(Object obsoletePos: pos.getObsolete()) {
			PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getTarget(), callInfo.getParent().getTargetProperty());
			originalAction = new RemoveElementAction(propertyKey, collectionElement, collectionOwner, null, obsoletePos, callInfo);
			this.addAction(originalAction);		
		}

		// Set the backRef value to null. This is an asynchronous activity
		if(originalAction != null && callInfo.getParent().getTargetProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getTargetProperty().getOpposite());			
			callInfo.getTargetRoot().getDependencyGraph().addAction(new SetterAction(null, newOppositeKey, originalAction));
		}
	}	

	@Override
	public void removeObsoletes(CallInfo callInfo) {	
		this.callInfo = callInfo;
		trackPosition(callInfo);

		super.removeObsoletes(callInfo);
	}

	@Override
	public void addOrReposition(CallInfo callInfo) throws Exception {	


		Map<Object, ExtendedDataObject> targetKeys = new HashMap<Object, ExtendedDataObject>();
		for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getTarget()).getList(callInfo))
			targetKeys.put(element.getIdentifierValue(), element);

		if(callInfo.getSource() != null) {
			AbstractOperation oper = (AbstractOperation) callInfo.getOperation();	
			for(ExtendedDataObject childDO: ((ExtendedDataObject)callInfo.getSource()).getList(callInfo)) {
				Position position = positions.get(childDO.getIdentifierValue());

				ExtendedDataObject targetElement = targetKeys.get(childDO.getIdentifierValue());
				CallInfo valueCI = new CallInfo(childDO, null, callInfo, null);
				if(callInfo.getSettings().getAction() != AggregateAction.CREATE && targetElement != null) { // updating an existing object

					valueCI.setTarget(targetElement);
					oper.processValue(valueCI);

				} else {

					valueCI.setTarget(oper.createTarget(valueCI, null));
					oper.processValue(valueCI);

					// AddElementAction
					addMapElement(valueCI, (ExtendedDataObject) valueCI.getTarget(), position);
				}
				// Add edge in dependency graph
				AbstractOperation.addEdge(valueCI);				
			}	
		}
	}

	protected void repositionElement(CallInfo callInfo, ExtendedDataObject collectionElement, Position position) {
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getTarget();
		
		// We need to add references for new map keys
		RepositionElementAction originalAction = null;
		for(Object addedPos: position.getAdded()) {
			PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getTarget(), callInfo.getParent().getTargetProperty());
			originalAction = new RepositionElementAction(propertyKey, collectionElement, collectionOwner, null, addedPos, callInfo);
			this.addAction(originalAction);
		}

		// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
		// TODO: handle MANY_TO_MANY backRef
		if(originalAction != null && callInfo.getParent().getTargetProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getTargetProperty().getOpposite());			
			callInfo.getTargetRoot().getDependencyGraph().addAction(new SetterAction(collectionElement, newOppositeKey, originalAction));
		}

	}

	protected void addMapElement(CallInfo callInfo, ExtendedDataObject collectionElement, Position position) {
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getTarget();

		if(position.isNew()) {
			AddElementAction originalAction = null;
			for(Object pos: position.getAdded()) {
				PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getTarget(), callInfo.getParent().getTargetProperty());
				originalAction = new AddElementAction(propertyKey, collectionElement, collectionOwner, null, pos, callInfo);
				this.addAction(originalAction);
			}

			// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
			// TODO: handle MANY_TO_MANY backRef			
			if(originalAction != null && callInfo.getParent().getTargetProperty().isBiDirectional()) {
				PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getTargetProperty().getOpposite());			
				callInfo.getTargetRoot().getDependencyGraph().addAction(new SetterAction(collectionElement, newOppositeKey, originalAction));
			}
		} else 
			repositionElement(callInfo, collectionElement, position);			
	}		
}
