/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model.operation;

import javax.persistence.metamodel.Attribute.PersistentAttributeType;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.exception.PropertyMappingException;
import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.ProcessingStage;
import ws.ladder.model.action.AddElementAction;
import ws.ladder.model.action.CollectionUpdateAction;
import ws.ladder.model.action.Executable;
import ws.ladder.model.action.PropertyKey;
import ws.ladder.model.action.RemoveElementAction;
import ws.ladder.model.action.SetterAction;


public class ModifyOperation extends AbstractOperation {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private Object result;

	public void removeObsoleteEntities(CallInfo callInfo, CollectionUpdateAction migratorAction) {

		// If we are merging then obsolete elements are not removed 		
		if(callInfo.getSettings().doMerge(callInfo))
			return;		

		migratorAction.removeObsoletes(callInfo);

	}	

	@Override
	public void processCollection(CallInfo callInfo) throws Exception {

		if(!callInfo.isCascadable()) // Hibernate will not update this collection
			return;

		if(callInfo.getStage() != ProcessingStage.UPDATE)
			return;		

		updateToMany(callInfo, null);
	}		

	protected void updateToMany(CallInfo callInfo, Executable triggeringAction) throws Exception {
		// Update the collection/map. This includes adding/removing/repositioning elements
		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getTarget(), callInfo.getTargetProperty());
		CollectionUpdateAction migratorAction = callInfo.getTargetRoot().getDependencyGraph().getOrCreateMigratorAction(propertyKey);

		removeObsoleteEntities(callInfo, migratorAction);
		migratorAction.addOrReposition(callInfo);
	}

	@Override
	protected boolean shouldUnlink(CallInfo ci) {

		if(ci.getSource() == null || ci.getSourceProperty() == null)
			return true;

		return ci.getSourceProperty().isTransient();
	}	

	@Override
	protected void processNullValue(CallInfo ci) throws Exception {

		if(ci.getSettings().doUpdate(ci) && ci.getStage() == ProcessingStage.UPDATE) {

			if(ci.getSourceProperty() != null && ci.getSourceProperty().isMany()) {
				ci.setTarget(createTarget(ci, ci.getTargetFromParent(), null)); // Get the persistent data object into target
				processCollection(ci);
			} else
				unlinkToOne(ci);
		}
	}	

	protected void unlinkToOne(CallInfo ci) {
		if(ci.getTargetProperty().getAssociationType() != PersistentAttributeType.ONE_TO_ONE &&
				ci.getTargetProperty().getAssociationType() != PersistentAttributeType.MANY_TO_ONE)
			return;

		if(ci.getTargetFromParent() == null)
			return;		

		// Set the target to null
		PropertyKey key = new PropertyKey((ExtendedDataObject)ci.getParentTargetEntity(), ci.getTargetProperty());		
		Executable originalAction = new SetterAction(null, key, null);
		ci.getTargetRoot().getDependencyGraph().addAction(originalAction);		

		// bi-directional link
		if(ci.getTargetProperty().isBiDirectional()) {

			ci.setTarget(createTarget(ci, ci.getTargetFromParent(), null)); // Get the persistent data object into target
			PropertyKey oppositeKey = new PropertyKey((ExtendedDataObject)ci.getTarget(), ci.getTargetProperty().getOpposite());

			// set the backRef from the opposite object to null	
			Executable oppositeAction = new SetterAction(null, oppositeKey, originalAction);
			if(ci.getTargetProperty().getAssociationType() == PersistentAttributeType.MANY_TO_ONE) {
				Object position = (oppositeKey.getProperty() == null) ? null : ((ExtendedProperty)oppositeKey.getProperty()).getValue(key.getDataObject());
				oppositeAction = new RemoveElementAction(oppositeKey, key.getDataObject(), (ExtendedDataObject) ci.getTarget(), originalAction, position, ci);				
			}
			ci.getTargetRoot().getDependencyGraph().addAction(oppositeAction);			
		}
	}

	private boolean isSameTarget(CallInfo ci, Object value) {
		return ci.getTargetFromParent() == ((ExtendedDataObject)value).getInstance();
	}

	protected void linkToOne(CallInfo ci, Object value) {
		if(ci.getTargetProperty().getAssociationType() != PersistentAttributeType.ONE_TO_ONE &&
				ci.getTargetProperty().getAssociationType() != PersistentAttributeType.MANY_TO_ONE)
			return;

		// If this is a uni-directional OneToOne, we just have to set the target to value
		PropertyKey key            = new PropertyKey((ExtendedDataObject)ci.getParentTargetEntity(), ci.getTargetProperty());
		Executable originalAction  = new SetterAction(value, key, null);

		// Set the target to value
		if(!isSameTarget(ci, value))		
			ci.getTargetRoot().getDependencyGraph().addAction(originalAction);			

		if(ci.getTargetProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey((ExtendedDataObject)value, ci.getTargetProperty().getOpposite());	

			// If the target has an existing opposite object, its backRef needs to be set to null 
			if(ci.getTargetFromParent() != null && !isSameTarget(ci, value)) {
				CallInfo newCall = new CallInfo(null, null, ci, ci.getTargetProperty());
				ExtendedDataObject oldObject = (ExtendedDataObject) createTarget(newCall, ci.getTargetFromParent(), null); // Get the persistent data object into target

				// Set the backRef of the opposite object to null
				PropertyKey oldOppositeKey = new PropertyKey(oldObject, ci.getTargetProperty().getOpposite());

				// Set the backRef from the old value to null
				Executable oppositeAction = new SetterAction(null, oldOppositeKey, originalAction);				
				if(ci.getTargetProperty().getAssociationType() == PersistentAttributeType.MANY_TO_ONE) {
					Object position = (oldOppositeKey.getProperty() == null) ? null : ((ExtendedProperty)oldOppositeKey.getProperty()).getValue(key.getDataObject());
					oppositeAction = new RemoveElementAction(oldOppositeKey, key.getDataObject(), oldObject, originalAction, position, ci);				
				}
				ci.getTargetRoot().getDependencyGraph().addAction(oppositeAction);	
			}			 

			// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
			Executable oppositeAction = new SetterAction(ci.getParentTargetEntity(), newOppositeKey, originalAction);			
			if(ci.getTargetProperty().getAssociationType() == PersistentAttributeType.MANY_TO_ONE) {
				Object position = (newOppositeKey. getProperty() == null) ? null : ((ExtendedProperty)newOppositeKey.getProperty()).getValue(key.getDataObject());
				oppositeAction = new AddElementAction(newOppositeKey, key.getDataObject(), (ExtendedDataObject) value, originalAction, position, ci);				
			}
			ci.getTargetRoot().getDependencyGraph().addAction(oppositeAction);		
		}
	}

	protected Object setPropertyTarget(CallInfo ci, Object propertyTarget) throws PropertyMappingException {
		if(ci.getStage() != ProcessingStage.UPDATE)
			return propertyTarget;

		// TODO: If the property is a mapKey or an indexKey then the toMany side needs to know about the reposition action 

		if(ci.getTargetProperty().isMany())
			super.setPropertyTarget(ci, propertyTarget);
		else 
			linkToOne(ci, propertyTarget);

		return propertyTarget;
	}

	@Override
	protected void setResult(Object target) {
		this.result = target;
	}	

	@Override
	public Object getResult() {
		return result;
	}	
}
