/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model.view;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.xml.bind.annotation.XmlRootElement;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.Type;
import ws.ladder.model.view.ViewBranch.ViewKey;

@XmlRootElement(name="ContentView")
public class ContentView {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	// The user can specify child branches if necessary directly. 
	// NOTE: If one or more child branch are specified, then only the child branches are executed.
	protected String            name;
	protected List<String>      attributeList;
	protected List<Filter>      filter;      // Each filter item should be independent of one another
	// It should refer to the full attribute name
	protected NativeQuery       nativeQuery; // SQL - the column names should be PROP1, PROP2 etc...
	//       validation should be done by the PersistenceOrchestrator
	protected List<ContentView> children;    // If the children are specified, it is not necessary to split it automatically
	// and we will trust the user has done a better split
	// NOTE: Only one level of child Content views are permitted. Nesting is not allowed.
	protected List<Parameter>   parameter;   // The parameters specified in the filter or in the native query

	public List<Parameter> getParameter() {
		return parameter;
	}

	public void setParameter(List<Parameter> parameter) {
		this.parameter = parameter;
	}

	public List<ContentView> getChildren() {
		return children;
	}

	public void setChildren(List<ContentView> children) {
		this.children = children;
	}

	public List<Filter> getFilter() {
		return filter;
	}

	public void setFilter(List<Filter> filters) {
		this.filter = filters;
	}	

	public NativeQuery getNativeQuery() {
		return nativeQuery;
	}

	public void setNativeQuery(NativeQuery nativeQuery) {
		this.nativeQuery = nativeQuery;
	}	

	public List<String> getAttributeList() {
		return attributeList;
	}

	public void setAttributeList(List<String> attributeList) {
		this.attributeList = attributeList;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public boolean hasNativeQuery() {
		boolean result = true;

		if(children == null)
			result = nativeQuery != null;
		else {
			boolean found = false;
			for(ContentView child: children) {
				if(!child.hasNativeQuery())
					result = false;
				else
					found = true;
			}
			if(found && !result)
				logger.warn("If a native query is specified, it needs to be specified on all branches or it will not be used");
		}

		return result;
	}

	public ViewBranch getEntityView(Type type, boolean narrow) {
		if(children != null) {
			for(ContentView child: children)
				if(child.children.size() > 0)
					throw new IllegalArgumentException("A content view cannot have more than one level of child content views");
		}
		ViewBranch result = ViewBranch.get(new ViewKey(type, name, narrow), this);

		return result;
	}

	public Set<String> getParameterNames() {
		Set<String> result = new HashSet<String>();

		if(parameter != null)
			for(Parameter param: parameter)
				result.add(param.name);

		return result;
	}
}