/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.service;

import java.util.ArrayList;
import java.util.List;

import javax.annotation.PostConstruct;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import ws.ladder.core.Interceptor;
import ws.ladder.model.AggregateAction;
import ws.ladder.model.DefaultTypeMapper;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedType;
import ws.ladder.model.MapperDirection;
import ws.ladder.model.Settings;
import ws.ladder.model.TypeMapper;
import ws.ladder.model.custom.AssociationStrategy;
import ws.ladder.model.custom.DefaultAssociationStrategy;
import ws.ladder.model.view.ContentViewFactory;
import ws.ladder.util.AccessType;
import ws.ladder.util.ObjectCreator;

@Component
public class AggregateManager {

	@Autowired
	protected SpringDASFactory springDASFactory;	

	// Configuration information
	protected List<String>        viewFiles;           // A list of files to read the view information from
	protected MetaModel           metaModel;           // Meta model exposed to the user
	protected boolean             crossAggregate;      // A system wide setting to allow cross aggregate access
	protected boolean             autoFlushNative;     // Flag to indicate if a flush should be issued before a native query is executed
	protected Interceptor         interceptor;         // A user provided interceptor to be notified of framework events
	protected AssociationStrategy associationStrategy;
	protected boolean             generateMetaModel;   // Generated metamodel on startup. This allows content integrators to find out the aggregate properties to be used in the views.
	protected TypeMapper          customTypeMapper;    // Custom type mapper to map between differnt types. NOTE: The user is restricted from changing the path specified in the MetaModel.
	// A truly custom solution is to provide a map between the meta model path and the custom path.
	// Configured with the DefaultTypeMapper
	protected AccessType          accessType;

	@PostConstruct
	protected void init() {
		// Load the default view file
		(new ContentViewFactory()).load();

		// Load the configured view files
		syncViews();
		
		if(associationStrategy == null)
			associationStrategy = new DefaultAssociationStrategy();
		
		if(customTypeMapper == null)
			customTypeMapper = new DefaultTypeMapper();
	}

	public void syncViews() {
		if(viewFiles == null)
			return;

		// Load the views
		ContentViewFactory viewFactory = new ContentViewFactory();

		for(String viewFile: viewFiles)
			viewFactory.load(viewFile);		
	}
	
	public AssociationStrategy getAssociationStrategy() {
		return associationStrategy;
	}

	public void setAssociationStrategy(AssociationStrategy associationStrategy) {
		this.associationStrategy = associationStrategy;
	}	
	
	public AccessType getAccessType() {
		return accessType;
	}

	public void setAccessType(AccessType accessType) {
		this.accessType = accessType;
	}	

	public List<String> getViewFiles() {
		return viewFiles;
	}

	public void setViewFiles(List<String> viewFiles) {
		this.viewFiles = viewFiles;
	}

	public boolean isCrossAggregate() {
		return crossAggregate;
	}

	public void setCrossAggregate(boolean crossAggregate) {
		this.crossAggregate = crossAggregate;
	}

	public boolean isAutoFlushNative() {
		return autoFlushNative;
	}

	public void setAutoFlushNative(boolean autoFlushNative) {
		this.autoFlushNative = autoFlushNative;
	}

	public Interceptor getInterceptor() {
		return interceptor;
	}

	public void setInterceptor(Interceptor interceptor) {
		this.interceptor = interceptor;
	}

	public boolean isGenerateMetaModel() {
		return generateMetaModel;
	}

	public void setGenerateMetaModel(boolean generateMetaModel) {
		this.generateMetaModel = generateMetaModel;
	}

	public TypeMapper getCustomTypeMapper() {
		return customTypeMapper;
	}

	public void setCustomTypeMapper(TypeMapper customTypeMapper) {
		this.customTypeMapper = customTypeMapper;
	}	

	protected DataAccessService getDAS() {
		DataAccessService das = springDASFactory.create(); // TODO: For now we create this object for every call, this can later be optimized

		return das;
	}
	
	public Settings getSettings() {
		Settings result = new Settings();
		result.setAssociationStrategy(associationStrategy);
		
		return result;
	}
	
	private void checkAndSet(Settings settings) {
		if(settings.getAssociationStrategy() == null)
			settings.setAssociationStrategy(associationStrategy);
	}

	public Object clone(Object entity, Class<?> desiredSubClass) {
		DataAccessService das = getDAS();

		ObjectCreator oc = new ObjectCreator(das, MapperDirection.REFERENCETOREFERENCE);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) das.getType(entity.getClass()), null, null);
		ExtendedDataObject to = (ExtendedDataObject) from.clone(getSettings());		

		return to.getInstance();
	}

	public void delete(Object entity) {
		// TODO: 
	}	

	public List<?> query(Object entity, Settings settings) {
		checkAndSet(settings);
		
		DataAccessService das = getDAS();
		if(settings.doPreFlush())
			das.getPersistenceOrchestrator().flush();

		ObjectCreator oc = new ObjectCreator(das, MapperDirection.REFERENCETODERIVED);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) das.getType(entity.getClass()), null, null);		

		List<?> dataObjects = from.query(settings);
		List<Object> result = new ArrayList<Object>();
		for(Object obj: dataObjects) 
			result.add( ((ExtendedDataObject)obj).getInstance() );

		return result;
	}

	public Object read(Object entity, Settings settings) {
		checkAndSet(settings);		
		DataAccessService das = getDAS();

		ObjectCreator oc = new ObjectCreator(das, MapperDirection.REFERENCETODERIVED);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) das.getType(entity.getClass()), null, null);
		ExtendedDataObject to = (ExtendedDataObject) from.read(settings);		

		return to.getInstance();
	} 	

	public Object create(Object entity, Settings settings) {
		checkAndSet(settings);		
		DataAccessService das = getDAS();

		ObjectCreator oc = new ObjectCreator(das, MapperDirection.DERIVEDTOREFERENCE);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) das.getType(entity.getClass()), null, null);
		ExtendedDataObject to = (ExtendedDataObject) from.create(settings);					

		return to.getInstance();		
	}

	public Object update(Object entity, Settings settings) {
		checkAndSet(settings);		
		DataAccessService das = getDAS();

		if(settings.getAction() != AggregateAction.MERGE && settings.getAction() != AggregateAction.UPDATE)
			throw new IllegalStateException("The default action should either by UPDATE or MERGE");

		ObjectCreator oc = new ObjectCreator(das, MapperDirection.DERIVEDTOREFERENCE);
		ExtendedDataObject from = oc.createDataObject(entity, (ExtendedType) das.getType(entity.getClass()), null, null);
		ExtendedDataObject to = (ExtendedDataObject) from.update(settings);	
		
		if(settings.doPostFlush())
			das.getPersistenceOrchestrator().flush();

		return to.getInstance();
	}	

}
