/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.service;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.BeanCreationException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.AutowireCapableBeanFactory;
import org.springframework.stereotype.Component;

import ws.ladder.model.MapperDirection;
import ws.ladder.model.TypeMapper;
import ws.ladder.util.AccessType;

@Component
public class SpringDASFactory implements DASFactory {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	@Autowired
	protected AutowireCapableBeanFactory beanFactory;

	@Autowired
	protected AggregateManager aggregateManager;
	
	protected DataAccessService das; // performance optimization

	public DataAccessService create() {
		if(das != null) 
			return das;
		
		// Needed for creating the derived types
		TypeMapper typeMapper = aggregateManager.getCustomTypeMapper().newInstance(MapperDirection.REFERENCETODERIVED);
		AccessType accessType = aggregateManager.getAccessType(); // If null, then we get the first working version

		try { // HibernateDAS
			if(accessType == null || accessType == AccessType.HIBERNATE) {
				PersistenceOrchestrator persistenceOrchestrator = new HibernatePersistenceOrchestrator();
				das = new HibernateDAS(typeMapper);
				beanFactory.autowireBean(das);
				beanFactory.autowireBean(persistenceOrchestrator);
				das.setPersistenceOrchestrator(persistenceOrchestrator);
				das.define();
				return das;
			}
		} catch(BeanCreationException e) {
			logger.warn("Hibernate configuration not found, hence cannot create a HibernateDAS instance");			
		}

		// This solution is not yet ready - should be ready in Hibernate 5
		try { // Hibernate configured using MetadataSources and ServiceRegistry
			if(accessType == null || accessType == AccessType.METADATA) {
				PersistenceOrchestrator persistenceOrchestrator = new HibernatePersistenceOrchestrator();				
				das = new MetadataDAS(typeMapper);
				beanFactory.autowireBean(das);
				beanFactory.autowireBean(persistenceOrchestrator);
				das.setPersistenceOrchestrator(persistenceOrchestrator);
				das.define();
				return das;		
			}
		} catch (BeanCreationException e) {
			logger.warn("Hibernate 5.0+ configuration not found, hence cannot create a MetadataDAS instance");
		}			

		try { // JPADataAccessService
			if(accessType == null || accessType == AccessType.JPA) {
				PersistenceOrchestrator persistenceOrchestrator = new JPAPersistenceOrchestrator();				
				das = new JPADataAccessService(typeMapper);
				beanFactory.autowireBean(das);
				beanFactory.autowireBean(persistenceOrchestrator);
				das.setPersistenceOrchestrator(persistenceOrchestrator);
				das.define();
				return das;		
			}
		} catch (BeanCreationException e) {
			logger.warn("JPA configuration not found, hence cannot create a JPADataAccessService instance");
		}

		return null;
	}

}
