/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.Property;
import ws.ladder.model.Settings;
import ws.ladder.model.SimpleType;
import ws.ladder.model.Type;

/**
 * Uses Brzozowski algebraic method to convert a Deterministic Finite Automata to a regular expression.
 * @author Dilip Dalton
 *
 */
public class DFAtoRE {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private Type                  aggregateType;
	private Map<Type, State>      states = new HashMap<Type, State>();
	private Stack<State>          navigationPath = new IdentityStack<State>();	
	private Map<Type, Expression> equations = new HashMap<Type, DFAtoRE.Expression>();
	private Map<State, Expression> regEx = new HashMap<DFAtoRE.State, DFAtoRE.Expression>();

	private String getFirstComponent(String propertyPath) {
		if(propertyPath != null && propertyPath.contains(Settings.PATH_DELIMITER))
			return propertyPath.substring(0, propertyPath.indexOf(Settings.PATH_DELIMITER));

		return propertyPath;
	}

	public boolean isValid(Type aggregate, String propertyPath) {
		String property = getFirstComponent(propertyPath);
		Type type = aggregate;

		while(property != null) {
			if(!isValidProperty(type, propertyPath))
				return false;

			type = states.get(type).getOutgoingTransition(property).end.type;		 
			propertyPath = getSubPath(propertyPath);
			property = getFirstComponent(propertyPath);
		}

		return true;
	}

	public String getSubPath(String propertyPath) {
		if(propertyPath.contains(Settings.PATH_DELIMITER))
			return propertyPath.substring(0, propertyPath.indexOf(Settings.PATH_DELIMITER));
		else
			return null;
	}

	/**
	 * Return the remaining path taking away the first component after its presence is checked in the type
	 * @param type
	 * @param property
	 * @return
	 */
	private boolean isValidProperty(Type type, String property) {
		return states.get(type).containsProperty(property);
	}

	public static class State {
		private Type type;
		private Map<String, Transition> outgoing = new HashMap<String, DFAtoRE.Transition>();
		private Set<Transition> incoming = new HashSet<DFAtoRE.Transition>();
		private boolean startState;
		private boolean finishState;
		private boolean inScope;

		public State(Type type, boolean startState) {
			this.type = type;
			this.setStartState(startState);
		}
		
		public String getName() {
			return this.type.getName();
		}
		
		public Type getType() {
			return this.type;
		}

		public void addOutgoingTransition(Transition transition) {
			if(transition.start.type != type)
				throw new IllegalArgumentException("The start type of the transition is not the same as the type of the state");

			if(outgoing.containsKey(transition.name))
				throw new IllegalArgumentException("Cannot have more than one transition witht the same input. We are modelling only DFA and not NFA");

			outgoing.put(transition.name, transition);
			transition.end.addIncomingTransition(transition);
		}

		public void addIncomingTransition(Transition transition) {
			incoming.add(transition);
		}

		public Transition getOutgoingTransition(String property) {
			return outgoing.get(property);
		}

		public boolean containsProperty(String property) {
			return outgoing.containsKey(property);
		}

		@Override
		public int hashCode() {
			int result = 17;
			result = 37 * result + type.hashCode();			
			return result;
		}

		@Override
		public boolean equals(Object other) {
			State otherState = (State) other;
			return otherState.type == this.type;
		}

		public boolean isFinishState() {
			return finishState;
		}

		public void setFinishState(boolean finishState) {
			this.finishState = finishState;
		}

		public boolean isStartState() {
			return startState;
		}

		public void setStartState(boolean startState) {
			this.startState = startState;
		}		

		public Expression getExpression() {

			Expression result = null;
			Expression previous = null;
			for(Transition transition: outgoing.values()) {
				if(!transition.end.isInScope())
					continue;

				if(previous != null) { // We have more than one
					if(result == null)
						result = new UnionExpression(previous);
					else
						((UnionExpression)result).addAlternate(previous);
				}

				Expression exp = new LiteralExpression(transition);
				previous = new TypedExpression(exp, transition.end.type);
			}

			if(result != null || isFinishState()) {
				if(result == null) {
					if(previous != null)
						result = new UnionExpression(previous);
				} else
					((UnionExpression)result).addAlternate(previous);

				if(result != null) {
					if(isFinishState())
						((UnionExpression)result).addAlternate(LiteralExpression.EMPTY_STRING);

					if(((UnionExpression)result).children.size() == 1)
						result = ((UnionExpression)result).children.iterator().next();
					else
						result = ((UnionExpression)result).consolidateTypes();
				} else
					result = LiteralExpression.EMPTY_STRING;
			} else {
				if(isFinishState()) {
					if(previous != null) {
						result = new UnionExpression(previous);
						((UnionExpression)result).addAlternate(LiteralExpression.EMPTY_STRING);
						result = ((UnionExpression)result).consolidateTypes();
					} else
						return LiteralExpression.EMPTY_STRING;
				} else 
					result = previous;
			}

			return result;
		}

		public boolean isInScope() {
			return inScope;
		}

		public void setInScope(boolean inScope) {
			this.inScope = inScope;
		}

		public void scopeStart() {
			if(isInScope())
				return;

			setInScope(true);
			for(Transition transition: incoming)
				transition.start.scopeStart();
		}
	}

	public static final class Transition {

		private final String name;
		private final State start;
		private final State end;

		public Transition(String name, State start, State end) {
			this.name = name;
			this.start = start;
			this.end = end;
		}
	}

	/**
	 * An expression might consist of an operation or a literal 
	 *
	 */
	public static interface Expression {

		public Expression append(Expression expression);		

		public Expression prepend(Expression expression);	

		public Expression substitute(Type typeToRemove, Expression expression);	

		public Expression reduce(Type type);

		public int getNumUnknowns();

		public Expression copy(); // return a deep copy

		public int getNumExpressions();

		public boolean isRecursive(Type type);
	}

	public static abstract class AbstractExpression implements Expression {

		@Override
		public Expression substitute(Type typeToRemove, Expression expression) {
			return this;
		}

		@Override
		public Expression reduce(Type type) {
			return this;
		}

		@Override
		public int getNumUnknowns() {
			return 0;
		}

		@Override
		public boolean isRecursive(Type type) {
			return false;
		}
	}

	public static class ConcatExpression extends AbstractExpression {
		private List<Expression> children = new ArrayList<DFAtoRE.Expression>();

		public ConcatExpression(Expression expression) {
			children.add(expression);
		}

		public Expression append(Expression expression) {
			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				return this;

			if(TypedExpression.class.isAssignableFrom(expression.getClass())) {
				Expression result = expression.prepend(this);
				return result;
			} else if(UnionExpression.class.isAssignableFrom(expression.getClass())) {
				if(expression.getNumUnknowns() == 0) {
					children.add(expression);
					return this;
				} else {
					Expression result = expression.prepend(this);
					return result;
				}
			} else {
				children.add(expression);
				return this;
			}
		}

		public Expression prepend(Expression expression) {
			expression = expression.copy();
			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				return this;

			children.add(0, expression);
			return this;
		}		

		@Override
		public String toString() {
			StringBuffer result = new StringBuffer();
			for(Expression expression: children)
				result.append(expression.toString());

			return result.toString();			
		}

		@Override
		public Expression copy() {
			Iterator<Expression> iter = children.iterator();
			ConcatExpression result = new ConcatExpression(iter.next().copy());

			while(iter.hasNext()) 
				result.children.add(iter.next().copy());

			return result;
		}

		@Override
		public int getNumExpressions() {
			int result = 0;

			for(Expression child: children)
				result += child.getNumExpressions();

			return result;
		}
	}	

	public static class UnionExpression extends AbstractExpression {
		private Set<Expression> children = new HashSet<DFAtoRE.Expression>();

		public UnionExpression(Expression expression) {
			if(UnionExpression.class.isAssignableFrom(expression.getClass()))
				children.addAll( ((UnionExpression)expression).children );
			else
				children.add(expression);
		}

		public void addAlternate(Expression expression) {
			if(UnionExpression.class.isAssignableFrom(expression.getClass()))
				children.addAll( ((UnionExpression)expression).children );
			else
				children.add(expression);
		}

		public void setChildren(Set<Expression> newChildren) {
			children = newChildren;

			for(Expression child: children) {
				if(child == this)
					throw new IllegalArgumentException("Cannot recursively reference an expression");
			}
		}

		@Override
		public Expression substitute(Type typeToRemove, Expression expression) {
			expression = expression.copy();
			Set<Expression> result = new HashSet<DFAtoRE.Expression>();		

			for(Expression child: children) {
				Expression exp = child.substitute(typeToRemove, expression);
				if(UnionExpression.class.isAssignableFrom(exp.getClass()))
					result.addAll( ((UnionExpression)exp).children );
				else
					result.add(exp);
			}

			setChildren(result);
			return consolidateTypes();
		}

		private Set<Expression> getNonTyped() {
			Set<Expression> result = new HashSet<DFAtoRE.Expression>();

			for(Expression child: children)
				if(!TypedExpression.class.isAssignableFrom(child.getClass()))
					result.add(child);

			return result;
		}

		@Override
		public Expression reduce(Type type) {
			Set<Expression> rest = new HashSet<Expression>();
			Expression starExpression = null;
			for(Expression child: children) {
				if(TypedExpression.class.isAssignableFrom(child.getClass()) && ((TypedExpression)child).getType() == type) {
					starExpression = ((TypedExpression)child).getExpression();
					starExpression = new StarExpression(starExpression);
				} else
					rest.add(child);
			}

			if(starExpression != null)
				if(rest.size() == 0)
					return starExpression;
				else if(rest.size() == 1)
					return starExpression.append(rest.iterator().next());
				else {
					setChildren(rest);
					return starExpression.append(this);
				}
			else
				return this;
		}

		@Override
		public Expression append(Expression expression) {
			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				return this;

			Set<Expression> result = new HashSet<Expression>();
			for(Expression child: children) {
				Expression newExpression = child.append(expression);
				if(UnionExpression.class.isAssignableFrom(newExpression.getClass()))
					result.addAll( ((UnionExpression)newExpression).children );
				else
					result.add(newExpression);				
			}
			setChildren(result);
			return this;
		}			

		public Expression prepend(Expression expression) {
			expression = expression.copy();
			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				return this;

			Set<Expression> result = new HashSet<Expression>();
			for(Expression child: children) {
				Expression newExpression = child.prepend(expression);
				if(UnionExpression.class.isAssignableFrom(newExpression.getClass()))
					result.addAll( ((UnionExpression)newExpression).children );
				else
					result.add(newExpression);
			}
			setChildren(result);
			return this;
		}	

		// find all the expressions of the same type and create a union expression out of them
		public Expression consolidateTypes() {
			logger.debug("Before consolidation: " + this.toString());

			boolean needsConsolidation = false;
			Map<Type, Set<Expression>> expressionsByType = new HashMap<Type, Set<Expression>>();
			for(Expression child: children) {
				if(TypedExpression.class.isAssignableFrom(child.getClass())) {
					TypedExpression typedChild = (TypedExpression) child;
					Set<Expression> expressions = expressionsByType.get(typedChild.getType());
					if(expressions == null) {
						expressions = new HashSet<DFAtoRE.Expression>();
						expressionsByType.put(typedChild.getType(), expressions);
					}
					expressions.add(typedChild.getExpression());
					if(expressions.size() > 1)
						needsConsolidation = true;
				}
			}

			if(needsConsolidation) {
				Set<Expression> result = getNonTyped();
				for(Type type: expressionsByType.keySet()) {
					Set<Expression> expressions = expressionsByType.get(type);
					if(expressions.size() > 1) {
						// create UnionExpression with the appropriate type
						Iterator<Expression> iter = expressions.iterator();
						UnionExpression expression = new UnionExpression(iter.next());
						while(iter.hasNext()) 
							expression.addAlternate(iter.next());
						result.add(new TypedExpression(expression, type));
					} else
						result.add(new TypedExpression(expressions.iterator().next(), type));
				}
				setChildren(result);

				if(result.size() == 1)
					return result.iterator().next();
			}

			return this;
		}

		@Override
		public int getNumUnknowns() {
			// The types should have been consolidated before this method is invoked, otherwise we will get the wrong answer
			int result = 0;
			for(Expression child: children)
				result += child.getNumUnknowns();

			return result;
		}	

		@Override
		public String toString() {
			StringBuffer result = new StringBuffer();
			for(Expression expression: children) {
				if(result.length() > 0)
					result.append(" + ");
				result.append(expression.toString());
			}
			if(children.size() < 2)
				throw new IllegalStateException("Union expression should consist of 2 or more expressions");

			result.insert(0, "(");
			result.append(")");

			return result.toString();			
		}

		@Override
		public Expression copy() {
			Iterator<Expression> iter = children.iterator();
			UnionExpression result = new UnionExpression(iter.next().copy());

			while(iter.hasNext()) 
				result.children.add(iter.next().copy());

			return result;
		}	

		@Override
		public int getNumExpressions() {
			int result = 0;

			for(Expression child: children)
				result += child.getNumExpressions();

			return result;
		}	

		@Override
		public boolean isRecursive(Type type) {
			for(Expression child: children)
				if(child.isRecursive(type))
					return true;

			return false;
		}
	}

	public static class StarExpression extends AbstractExpression {
		private Expression expression;

		public StarExpression(Expression expression) {
			this.expression = expression;

			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				throw new IllegalArgumentException("star on an empty string is an empty string, do not wrap with a star expression");

			if(StarExpression.class.isAssignableFrom(expression.getClass()))
				throw new IllegalArgumentException("Does not make sense to wrap a star expression inside another star expression");
		}

		@Override
		public Expression prepend(Expression expression) {
			expression = expression.copy();
			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				return this;

			ConcatExpression concatExp = new ConcatExpression(this);
			concatExp.prepend(expression);
			return concatExp;
		}

		@Override
		public Expression append(Expression expression) {
			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				return this;

			expression = expression.copy();
			if(TypedExpression.class.isAssignableFrom(expression.getClass())) {
				ConcatExpression concatExp = new ConcatExpression(this.copy());
				((TypedExpression)expression).nonTyped = concatExp.append( ((TypedExpression)expression).getExpression() );
				return expression;				

			} else {
				ConcatExpression concatExp = new ConcatExpression(this);
				concatExp.append(expression);
				return concatExp;
			}
		}

		@Override
		public String toString() {
			StringBuffer result = new StringBuffer(expression.toString());

			if(ConcatExpression.class.isAssignableFrom(expression.getClass())) {
				result.insert(0, "(");
				result.append(")");
			}
			result.append("*");

			return result.toString();
		}

		@Override
		public Expression copy() {
			return new StarExpression(expression.copy());
		}		

		@Override
		public int getNumExpressions() {
			return expression.getNumExpressions() + 1;
		}			
	}

	public static class LiteralExpression extends AbstractExpression {
		public static final LiteralExpression EMPTY_STRING = new LiteralExpression();
		public static final String EPSILON = "\u025B";		

		private Transition transition;		

		private LiteralExpression() {
		}

		public LiteralExpression(Transition transition) {
			this.transition = transition;
		}

		@Override
		public Expression prepend(Expression expression) {
			expression = expression.copy();
			if(LiteralExpression.class.isAssignableFrom(expression.getClass())) {
				Expression result = this;
				if( (LiteralExpression)expression != LiteralExpression.EMPTY_STRING) {
					result = new ConcatExpression(expression);
					result.append(this);
				}
				return result;
			} else {
				if(this == LiteralExpression.EMPTY_STRING)
					return expression;
				else
					return expression.append(this);
			}
		}

		@Override
		public Expression append(Expression expression) {
			if(LiteralExpression.class.isAssignableFrom(expression.getClass()) && (LiteralExpression)expression == LiteralExpression.EMPTY_STRING)
				return this;

			return expression.prepend(this);
		}

		@Override
		public String toString() {
			return (this == LiteralExpression.EMPTY_STRING) ? LiteralExpression.EPSILON : transition.name;
		}

		@Override
		public Expression copy() {
			if(this == LiteralExpression.EMPTY_STRING)
				return LiteralExpression.EMPTY_STRING;
			return new LiteralExpression(transition);
		}

		@Override
		public int getNumExpressions() {
			return 1;
		}		
	}	

	public static class TypedExpression extends AbstractExpression {
		private Expression nonTyped;
		private Type type;

		public TypedExpression(Expression nonTyped, Type type) {
			this.nonTyped = nonTyped;
			this.type = type;

			if(TypedExpression.class.isAssignableFrom(nonTyped.getClass()))
				throw new IllegalArgumentException("The expression is expected to be non-typed");
		}

		public Type getType() {
			return type;
		}

		public Expression getExpression() {
			return nonTyped;
		}

		@Override
		// We are doing substitution
		public Expression append(Expression expression) {
			return nonTyped.append(expression);
		}

		@Override
		public Expression prepend(Expression expression) {
			expression = expression.copy();
			nonTyped = nonTyped.prepend(expression);
			return this;
		}

		@Override
		public Expression substitute(Type typeToRemove, Expression substituteExp) {
			if(this.type == typeToRemove) {
				if(TypedExpression.class.isAssignableFrom(substituteExp.getClass())) {
					ConcatExpression concatExp = new ConcatExpression(nonTyped.copy());
					concatExp.append( ((TypedExpression)substituteExp).getExpression() );
					nonTyped = concatExp;
					this.type = ((TypedExpression)substituteExp).type;
					return this;
				} else {
					nonTyped = nonTyped.append(substituteExp.copy());
					return nonTyped;
				}
			}

			return this;
		}	

		@Override
		public int getNumUnknowns() {
			return 1;
		}		

		@Override
		public String toString() {
			StringBuffer result = new StringBuffer(nonTyped.toString());
			result.append(type.getName());

			return result.toString();
		}

		@Override
		public Expression copy() {
			return new TypedExpression(nonTyped.copy(), type);
		}		

		@Override
		public int getNumExpressions() {
			return nonTyped.getNumExpressions() + 1;
		}		

		@Override
		public boolean isRecursive(Type type) {
			if(this.type == type)
				return true;

			return false;
		}

		@Override
		public Expression reduce(Type type) {
			if(this.type == type)
				return new StarExpression( nonTyped );

			return this;
		}			
	}

	public void setAggregateType(Type type) {
		this.aggregateType = type;
	}

	public void addState(State state) {
		states.put(state.type, state);
	}

	public DFAtoRE() {
	}

	public DFAtoRE(Type aggregateType) {
		this.aggregateType = aggregateType;
		buildDFA();
		solve();
	}

	public void solve() {
		regEx = new HashMap<DFAtoRE.State, DFAtoRE.Expression>();
		
		int numStartStates = 0;
		for(State state: states.values()) {
			if(state.isStartState())
				numStartStates++;
		}
		if(numStartStates != 1) 
			throw new IllegalArgumentException("There should be only one start state");

		for(State state: states.values()) {
			state.setFinishState(true);
			createEquations();
			Expression result = processEquations();
			regEx.put(state, result);
			//System.out.println("finish type " + state.type.getName() + " : " + result.toString());

			state.setFinishState(false);
		}		
	}
	
	public Map<DFAtoRE.State, DFAtoRE.Expression> getRegEx() {
		return (Map<State, Expression>) Collections.unmodifiableMap(regEx);
	}

	private void processReachability() {
		State finishState = null;
		for(State state: states.values()) {
			if(state.isFinishState())
				finishState = state;
			state.setInScope(false);
		}

		finishState.scopeStart();
	}

	public Type getPropertyType(Property property) {
		Type result = property.getType();
		if(SimpleType.class.isAssignableFrom(result.getClass())) {
			if(property != null && property.isMany())
				result = ((ExtendedProperty)property).getElementType();
		}			

		return result;
	}	

	private void buildDFA() {
		// build the set of states and transitions for the aggregate type
		navigationPath.push(new State(aggregateType, true));
		execute();
	}

	protected void execute() {

		while(!navigationPath.empty()) {
			// Process the state object
			State state = navigationPath.pop();
			Type type = state.type;
			if(states.containsKey(type))
				throw new IllegalArgumentException("State with type name " + type.getName() + " is being processed multiple times");
			states.put(type, state);

			for(Property childProperty: type.getProperties()) {
				Type propertyType = getPropertyType(childProperty);

				// Don't walk through associations, allow SimpleType
				if(childProperty != null && !childProperty.isContainment() && !SimpleType.class.isAssignableFrom(aggregateType.getClass()))
					continue;			

				// check if it has been processed
				if(!propertyType.isDataType()) {		
					State endState = states.get(propertyType);
					if(!states.containsKey(propertyType)) {
						endState = new State(propertyType, false); 
						if(!navigationPath.contains(endState))
							navigationPath.push(endState);						
					}
					// add the transition
					Transition transition = new Transition(childProperty.getName()+Settings.PATH_DELIMITER, state, endState);
					state.addOutgoingTransition(transition);
				}
			}		
		}
	}

	public void createEquations() {
		equations = new HashMap<Type, DFAtoRE.Expression>();

		processReachability();
		for(State state: states.values()) {
			if(!state.isInScope())
				continue;

			Expression exp = state.getExpression();
			if(exp == null)
				continue;
			logger.debug("Equation for type " + state.type.getName() + ": " + exp.toString());

			equations.put(state.type, exp);
		}
	}

	private Type getTypeToRemove() {
		// Keep the logic simple
		// We choose the type with the minimum number of unknowns

		/*
		// Call reduce on all expressions except the starting state
		for(Type type: equations.keySet()) {
			State state = states.get(type);
			if(state.isStartState()) 
				continue;

			Expression exp = equations.get(type);
			Expression reduced = exp.reduce(type);
			equations.put(type, reduced);
		}*/

		int minUnknowns = Integer.MAX_VALUE;
		Type typeWithMinUnknowns = null;
		for(Type type: equations.keySet()) {
			State state = states.get(type);
			if(state.isStartState()) // We don't replace the start state, since that is the state we are interested in
				continue;

			Expression exp = equations.get(type);
			if(exp.getNumExpressions() < minUnknowns) {
				minUnknowns = exp.getNumExpressions();
				typeWithMinUnknowns = type;
			}
		}

		// reduce the recursive expression
		if(typeWithMinUnknowns != null) {
			Expression exp = equations.get(typeWithMinUnknowns);
			if(exp.isRecursive(typeWithMinUnknowns)) {
				Expression reducedExp = exp.reduce(typeWithMinUnknowns);
				logger.debug("reducing type " + typeWithMinUnknowns.getName() + " : " + reducedExp.toString());
				equations.put(typeWithMinUnknowns, reducedExp);
			}
		}

		return typeWithMinUnknowns;
	}

	private void substitute(Type typeToRemove) {
		for(Type type: equations.keySet()) {
			if(type == typeToRemove)
				continue;
			Expression exp = equations.get(type);
			equations.put(type, exp.substitute(typeToRemove, equations.get(typeToRemove)));
		}
	}

	private void printEquations(String message, Type typeToRemove) {
		if(typeToRemove == null)
			return;

		logger.debug(message + typeToRemove.getName());
		for(Type type: equations.keySet()) {
			logger.debug(type.getName() + " = " + equations.get(type).toString());
		}
	}

	public Expression processEquations() {
		Type typeToRemove = getTypeToRemove();
		printEquations("After reduction: ", typeToRemove);
		while(typeToRemove != null) {
			substitute(typeToRemove);
			printEquations("After substitution: ", typeToRemove);
			equations.remove(typeToRemove);			
			typeToRemove = getTypeToRemove();
		}

		assert(equations.size() == 1);
		assert(equations.containsKey(aggregateType));

		Expression result = equations.get(aggregateType);
		result = result.reduce(aggregateType);

		return result;
	}	
}